@@ -238,17 +238,25 @@ inline void enumerate_coarse_points(viennacl::linalg::detail::amg::amg_level_con
   unsigned int *point_types_ptr  = viennacl::linalg::host_based::detail::extract_raw_pointer<unsigned int>(amg_context.point_types_.handle());
   unsigned int *coarse_id_ptr    = viennacl::linalg::host_based::detail::extract_raw_pointer<unsigned int>(amg_context.coarse_id_.handle());
 
-  unsigned int coarse_id = 0;
+  unsigned int num_coarse = 0;
+#ifdef VIENNACL_WITH_OPENMP
+#pragma omp parallel for reduction(+: num_coarse)
+#endif
   for (vcl_size_t i=0; i<amg_context.coarse_id_.size(); ++i)
   {
-    //assert(point_types_ptr[i] != viennacl::linalg::detail::amg::amg_level_context::POINT_TYPE_UNDECIDED && bool("Logic error in enumerate_coarse_points(): Undecided points detected!"));
-
     if (point_types_ptr[i] == viennacl::linalg::detail::amg::amg_level_context::POINT_TYPE_COARSE)
-      coarse_id_ptr[i] = coarse_id++;
+    {
+      coarse_id_ptr[i] = 1;
+      num_coarse += 1;
+    }
+    else
+      coarse_id_ptr[i] = 0;
   }
 
-  //Rcpp::Rcout << "Coarse nodes after enumerate_coarse_points(): " << coarse_id << std::endl;
-  amg_context.num_coarse_ = coarse_id;
+  // possible improvement: fuse for-loop above with first stage of exclusive scan
+  viennacl::linalg::exclusive_scan(amg_context.coarse_id_, amg_context.coarse_id_);
+
+  amg_context.num_coarse_ = num_coarse;
 }
 
 
@@ -355,6 +363,168 @@ void amg_coarse_classic_onepass(compressed_matrix<NumericT> const & A,
 }
 
 
+/** @brief Classical coarsening using parallel maximum independent sets
+*
+* Algorithm follows PMIS as outlined in "Reducing Complexity in Parallel Algebraic Multigrid Preconditioners" by De Sterck, Meyer-Yang, Heys (2004)
+*
+* @param A             Operator matrix for the respective level
+* @param amg_context   AMG datastructure object for the grid hierarchy
+* @param tag           AMG preconditioner tag
+*/
+template<typename NumericT>
+void amg_coarse_classic_pmis1(compressed_matrix<NumericT> const & A,
+                              viennacl::linalg::detail::amg::amg_level_context & amg_context,
+                              viennacl::linalg::amg_tag & tag)
+{
+  (void)tag;
+
+  amg_influence_trivial(A, amg_context, tag);
+  //amg_influence_advanced(A, amg_context, tag);
+
+  unsigned int  *point_types_ptr       = viennacl::linalg::host_based::detail::extract_raw_pointer<unsigned int>(amg_context.point_types_.handle());
+  unsigned int *influences_row_ptr    = viennacl::linalg::host_based::detail::extract_raw_pointer<unsigned int>(amg_context.influence_jumper_.handle());
+  unsigned int *influences_id_ptr     = viennacl::linalg::host_based::detail::extract_raw_pointer<unsigned int>(amg_context.influence_ids_.handle());
+
+  std::vector<float> random_weights(A.size1());
+  for (std::size_t i=0; i<random_weights.size(); ++i)
+    random_weights[i] = float(influences_row_ptr[i+1] - influences_row_ptr[i]) + (float(rand()) / float(RAND_MAX));
+
+  std::size_t num_threads = 1;
+#ifdef VIENNACL_WITH_OPENMP
+  num_threads = omp_get_max_threads();
+#endif
+
+  viennacl::vector<float>        work_random2(A.size1(), viennacl::traits::context(A));
+  viennacl::vector<unsigned int> work_index2(A.size1(), viennacl::traits::context(A));
+
+  float        *work_random2_ptr    = viennacl::linalg::host_based::detail::extract_raw_pointer<float>(work_random2.handle());
+  unsigned int *work_index2_ptr     = viennacl::linalg::host_based::detail::extract_raw_pointer<unsigned int>(work_index2.handle());
+
+
+  unsigned int num_undecided = static_cast<unsigned int>(A.size1());
+  unsigned int pmis_iters = 0;
+
+  //
+  // Setup: Set orphaned points with no strong connection to fine points:
+  //
+  for (long i=0; i<static_cast<long>(A.size1()); ++i)
+  {
+    if (influences_row_ptr[i] == influences_row_ptr[i + 1])
+    {
+      point_types_ptr[i] = viennacl::linalg::detail::amg::amg_level_context::POINT_TYPE_FINE;
+    }
+  }
+
+
+  //
+  // Repeat PMIS iteration until all nodes have been decided
+  //
+  while (num_undecided > 0)
+  {
+    ++pmis_iters;
+
+    //
+    // Propagate maximum tuple once (only required for undecided points)
+    //
+#ifdef VIENNACL_WITH_OPENMP
+    #pragma omp parallel for
+#endif
+    for (long i=0; i<static_cast<long>(A.size1()); ++i)
+    {
+      if (point_types_ptr[i] != viennacl::linalg::detail::amg::amg_level_context::POINT_TYPE_UNDECIDED)
+        continue;
+
+      // load
+      float random = random_weights[i];
+      unsigned int index  = i;
+
+      // max
+      unsigned int j_stop = influences_row_ptr[i + 1];
+      for (unsigned int j = influences_row_ptr[i]; j < j_stop; ++j)
+      {
+        unsigned int influenced_point_id = influences_id_ptr[j];
+
+        if (point_types_ptr[influenced_point_id] != viennacl::linalg::detail::amg::amg_level_context::POINT_TYPE_UNDECIDED)
+          continue;
+
+        float other_random = random_weights[influenced_point_id];
+
+        if (    random < other_random
+            || (random <= other_random && index < influenced_point_id))
+        {
+          random = other_random;
+          index  = influenced_point_id;
+        }
+      } // for
+
+      // store
+      work_random2_ptr[i] = random;
+      work_index2_ptr[i]  = index;
+    }
+
+
+    //
+    // mark coarse nodes:
+    //
+
+// OpenMP acceleration for this loop disabled because of troubles with NVCC in CUDA 6.0.
+//#if defined(VIENNACL_WITH_OPENMP)
+//  #pragma omp parallel for
+//#endif
+    for (long i=0; i<static_cast<long>(A.size1()); ++i)
+    {
+      if (point_types_ptr[i] == viennacl::linalg::detail::amg::amg_level_context::POINT_TYPE_UNDECIDED)
+      {
+        //if (work_has_undecided_neighbor[i] && i == work_index2_ptr[i]) // this is a local maximum, so turn this into a coarse node
+        if (i == work_index2_ptr[i]) // this is a local maximum, so turn this into a coarse node
+        {
+          point_types_ptr[i] = viennacl::linalg::detail::amg::amg_level_context::POINT_TYPE_COARSE;
+
+          // set neighbors to fine nodes (race conditions are harmless here, because the same value is set to all strongly influenced undecided neighbor nodes)
+          unsigned int j_stop = influences_row_ptr[i + 1];
+          for (unsigned int j = influences_row_ptr[i]; j < j_stop; ++j)
+          {
+            unsigned int influenced_point_id = influences_id_ptr[j];
+            if (point_types_ptr[influenced_point_id] == viennacl::linalg::detail::amg::amg_level_context::POINT_TYPE_UNDECIDED)
+              point_types_ptr[influenced_point_id] = viennacl::linalg::detail::amg::amg_level_context::POINT_TYPE_FINE;
+          }
+        }
+      }
+    }
+
+    //
+    // count undecided nodes
+    //
+    std::vector<unsigned int> thread_buffer(num_threads);
+
+#ifdef VIENNACL_WITH_OPENMP
+    #pragma omp parallel for
+#endif
+    for (long i2=0; i2<static_cast<long>(A.size1()); ++i2)
+    {
+      unsigned int i = static_cast<unsigned int>(i2);
+
+      if (point_types_ptr[i] == viennacl::linalg::detail::amg::amg_level_context::POINT_TYPE_UNDECIDED)
+      {
+#ifdef VIENNACL_WITH_OPENMP
+        thread_buffer[omp_get_thread_num()] += 1;
+#else
+        thread_buffer[0] += 1;
+#endif
+      }
+    }
+
+    num_undecided = 0;
+    for (std::size_t i=0; i<thread_buffer.size(); ++i)
+      num_undecided += thread_buffer[i];
+  } // while
+
+  viennacl::linalg::host_based::amg::enumerate_coarse_points(amg_context);
+}
+
+
+
+
 //////////////////////////
 
 
@@ -408,6 +578,71 @@ void amg_coarse_ag_stage1_sequential(compressed_matrix<NumericT> const & A,
 }
 
 
+/** @brief Propagates the work state information a distance of two, using a maximum operation to overwrite 'smaller' work state */
+template<typename IndexT>
+void amg_coarse_ag_stage1_mis2_propagate(IndexT const *work_state_ptr,  IndexT const *work_random_ptr,  IndexT const *work_index_ptr,
+                                         IndexT       *work_state2_ptr, IndexT       *work_random2_ptr, IndexT       *work_index2_ptr,
+                                         IndexT const *influences_row_ptr, IndexT const *influences_id_ptr, long num_points)
+{
+
+#ifdef VIENNACL_WITH_OPENMP
+  #pragma omp parallel for
+#endif
+  for (long i=0; i<num_points; ++i)
+  {
+    // load
+    IndexT state  = work_state_ptr[i];
+    IndexT random = work_random_ptr[i];
+    IndexT index  = work_index_ptr[i];
+
+    // max
+    IndexT j_stop = influences_row_ptr[i + 1];
+    for (IndexT j = influences_row_ptr[i]; j < j_stop; ++j)
+    {
+      IndexT influenced_point_id = influences_id_ptr[j];
+
+      IndexT other_state = work_state_ptr[influenced_point_id];
+
+
+      // lexigraphical triple-max (not particularly pretty, but does the job):
+      if (state < other_state)
+      {
+        state  = other_state;
+        random = work_random_ptr[influenced_point_id];
+        index  = work_index_ptr[influenced_point_id];
+      }
+      else if (state == other_state)
+      {
+        IndexT other_random = work_random_ptr[influenced_point_id];
+
+        if (random < other_random)
+        {
+          state  = other_state;
+          random = other_random;
+          index  = work_index_ptr[influenced_point_id];
+        }
+        else if (random == other_random)
+        {
+          IndexT other_index = work_index_ptr[influenced_point_id];
+
+          if (index < other_index)
+          {
+            state  = other_state;
+            random = other_random;
+            index  = other_index;
+          }
+        }
+      }
+    } // for
+
+    // store
+    work_state2_ptr[i]  = state;
+    work_random2_ptr[i] = random;
+    work_index2_ptr[i]  = index;
+  }
+
+}
+
 
 /** @brief AG (aggregation based) coarsening, multi-threaded version of stage 1 using parallel maximum independent sets
 *
@@ -453,101 +688,47 @@ void amg_coarse_ag_stage1_mis2(compressed_matrix<NumericT> const & A,
 
   unsigned int num_undecided = static_cast<unsigned int>(A.size1());
   unsigned int pmis_iters = 0;
+
+  //
+  // init temporary work data:
+  //
+#ifdef VIENNACL_WITH_OPENMP
+  #pragma omp parallel for
+#endif
+  for (long i2=0; i2<static_cast<long>(A.size1()); ++i2)
+  {
+    unsigned int i = static_cast<unsigned int>(i2);
+    switch (point_types_ptr[i])
+    {
+    case viennacl::linalg::detail::amg::amg_level_context::POINT_TYPE_UNDECIDED: work_state_ptr[i] = 1; break;
+    case viennacl::linalg::detail::amg::amg_level_context::POINT_TYPE_FINE:      work_state_ptr[i] = 0; break;
+    case viennacl::linalg::detail::amg::amg_level_context::POINT_TYPE_COARSE:    work_state_ptr[i] = 2; break;
+    default:
+      throw std::runtime_error("Unexpected state encountered in MIS2 setup for AMG.");
+    }
+
+    //work_random_ptr[i] = random_weights[i]; // note: optimized out. Passing random_weights directly to first call to amg_coarse_ag_stage1_mis2_propagate()
+    work_index_ptr[i]  = i;
+  }
+
+
+  //
+  // Repeat PMIS iteration until all nodes have been decided
+  //
   while (num_undecided > 0)
   {
     ++pmis_iters;
 
-    //
-    // init temporary work data:
-    //
-#ifdef VIENNACL_WITH_OPENMP
-  #pragma omp parallel for
-#endif
-    for (long i2=0; i2<static_cast<long>(A.size1()); ++i2)
-    {
-      unsigned int i = static_cast<unsigned int>(i2);
-      switch (point_types_ptr[i])
-      {
-      case viennacl::linalg::detail::amg::amg_level_context::POINT_TYPE_UNDECIDED: work_state_ptr[i] = 1; break;
-      case viennacl::linalg::detail::amg::amg_level_context::POINT_TYPE_FINE:      work_state_ptr[i] = 0; break;
-      case viennacl::linalg::detail::amg::amg_level_context::POINT_TYPE_COARSE:    work_state_ptr[i] = 2; break;
-      default:
-        throw std::runtime_error("Unexpected state encountered in MIS2 setup for AMG.");
-      }
-
-      work_random_ptr[i] = random_weights[i];
-      work_index_ptr[i]  = i;
-    }
-
-
     //
     // Propagate maximum tuple twice
+    // (first call writes state from work_*_ptr to work_*2_ptr, second call from work_*2_ptr back to work_*_ptr)
     //
-    for (unsigned int r = 0; r < 2; ++r)
-    {
-      // max operation
-#ifdef VIENNACL_WITH_OPENMP
-      #pragma omp parallel for
-#endif
-      for (long i2=0; i2<static_cast<long>(A.size1()); ++i2)
-      {
-        unsigned int i = static_cast<unsigned int>(i2);
-        // load
-        unsigned int state  = work_state_ptr[i];
-        unsigned int random = work_random_ptr[i];
-        unsigned int index  = work_index_ptr[i];
-
-        // max
-        unsigned int j_stop = influences_row_ptr[i + 1];
-        for (unsigned int j = influences_row_ptr[i]; j < j_stop; ++j)
-        {
-          unsigned int influenced_point_id = influences_id_ptr[j];
-
-          // lexigraphical triple-max (not particularly pretty, but does the job):
-          if (state < work_state_ptr[influenced_point_id])
-          {
-            state  = work_state_ptr[influenced_point_id];
-            random = work_random_ptr[influenced_point_id];
-            index  = work_index_ptr[influenced_point_id];
-          }
-          else if (state == work_state_ptr[influenced_point_id])
-          {
-            if (random < work_random_ptr[influenced_point_id])
-            {
-              state  = work_state_ptr[influenced_point_id];
-              random = work_random_ptr[influenced_point_id];
-              index  = work_index_ptr[influenced_point_id];
-            }
-            else if (random == work_random_ptr[influenced_point_id])
-            {
-              if (index < work_index_ptr[influenced_point_id])
-              {
-                state  = work_state_ptr[influenced_point_id];
-                random = work_random_ptr[influenced_point_id];
-                index  = work_index_ptr[influenced_point_id];
-              }
-            } // max(random)
-          } // max(state)
-        } // for
-
-        // store
-        work_state2_ptr[i]  = state;
-        work_random2_ptr[i] = random;
-        work_index2_ptr[i]  = index;
-      }
-
-      // copy work array
-#ifdef VIENNACL_WITH_OPENMP
-      #pragma omp parallel for
-#endif
-      for (long i2=0; i2<static_cast<long>(A.size1()); ++i2)
-      {
-        unsigned int i = static_cast<unsigned int>(i2);
-        work_state_ptr[i]  = work_state2_ptr[i];
-        work_random_ptr[i] = work_random2_ptr[i];
-        work_index_ptr[i]  = work_index2_ptr[i];
-      }
-    }
+    amg_coarse_ag_stage1_mis2_propagate(work_state_ptr,  &(random_weights[0]),  work_index_ptr,
+                                        work_state2_ptr, work_random2_ptr, work_index2_ptr,
+                                        influences_row_ptr, influences_id_ptr, long(A.size1()));
+    amg_coarse_ag_stage1_mis2_propagate(work_state2_ptr, work_random2_ptr, work_index2_ptr,
+                                        work_state_ptr,  work_random_ptr,  work_index_ptr,
+                                        influences_row_ptr, influences_id_ptr, long(A.size1()));
 
     //
     // mark MIS and non-MIS nodes:
@@ -562,19 +743,36 @@ void amg_coarse_ag_stage1_mis2(compressed_matrix<NumericT> const & A,
       unsigned int i = static_cast<unsigned int>(i2);
       unsigned int max_state  = work_state_ptr[i];
       unsigned int max_index  = work_index_ptr[i];
+      work_index_ptr[i] = i; // reset for next iteration
 
-      if (point_types_ptr[i] == viennacl::linalg::detail::amg::amg_level_context::POINT_TYPE_UNDECIDED)
+      switch (point_types_ptr[i])
       {
+      case viennacl::linalg::detail::amg::amg_level_context::POINT_TYPE_UNDECIDED:
         if (i == max_index) // make this a MIS node
+        {
           point_types_ptr[i] = viennacl::linalg::detail::amg::amg_level_context::POINT_TYPE_COARSE;
+          work_state_ptr[i] = 2;
+        }
         else if (max_state == 2) // mind the mapping of viennacl::linalg::detail::amg::amg_level_context::POINT_TYPE_COARSE above!
+        {
           point_types_ptr[i] = viennacl::linalg::detail::amg::amg_level_context::POINT_TYPE_FINE;
+          work_state_ptr[i] = 0;
+        }
         else
+        {
+          work_state_ptr[i] = 1;
 #ifdef VIENNACL_WITH_OPENMP
           thread_buffer[omp_get_thread_num()] += 1;
 #else
           thread_buffer[0] += 1;
 #endif
+        }
+        break;
+
+      case viennacl::linalg::detail::amg::amg_level_context::POINT_TYPE_FINE:      work_state_ptr[i] = 0; break;
+      case viennacl::linalg::detail::amg::amg_level_context::POINT_TYPE_COARSE:    work_state_ptr[i] = 2; break;
+      default:
+        throw std::runtime_error("Unexpected state encountered in MIS2 setup for AMG.");
       }
     }
 
@@ -736,30 +934,27 @@ void amg_interpol_direct(compressed_matrix<NumericT> const & A,
   unsigned int *influences_id_ptr     = viennacl::linalg::host_based::detail::extract_raw_pointer<unsigned int>(amg_context.influence_ids_.handle());
   unsigned int *coarse_id_ptr         = viennacl::linalg::host_based::detail::extract_raw_pointer<unsigned int>(amg_context.coarse_id_.handle());
 
-  P.resize(A.size1(), amg_context.num_coarse_, false);
+  NumericT threshold = 1e-2;
 
-  std::vector<std::map<unsigned int, NumericT> > P_setup(A.size1());
+  //
+  // Step 1: Determine sparsity pattern of P:
+  //
+  unsigned int *P_nonzero_helper = (unsigned int*)malloc((A.size1()+1) * sizeof(unsigned int));
+  NumericT *interpolation_helper = (NumericT *)malloc(A.size1() * sizeof(NumericT));
 
-  // Iterate over all points to build the interpolation matrix row-by-row
-  // Interpolation for coarse points is immediate using '1'.
-  // Interpolation for fine points is set up via corresponding row weights (cf. Yang paper, p. 14)
 #ifdef VIENNACL_WITH_OPENMP
   #pragma omp parallel for
 #endif
   for (long row2=0; row2<static_cast<long>(A.size1()); ++row2)
   {
     unsigned int row = static_cast<unsigned int>(row2);
-    std::map<unsigned int, NumericT> & P_setup_row = P_setup[row];
-    //Rcpp::Rcout << "Row " << row << ": " << std::endl;
+
     if (point_types_ptr[row] == viennacl::linalg::detail::amg::amg_level_context::POINT_TYPE_COARSE)
     {
-      //Rcpp::Rcout << "  Setting value 1.0 at " << coarse_id_ptr[row] << std::endl;
-      P_setup_row[coarse_id_ptr[row]] = NumericT(1);
+      P_nonzero_helper[row] = 1;
     }
     else if (point_types_ptr[row] == viennacl::linalg::detail::amg::amg_level_context::POINT_TYPE_FINE)
     {
-      //Rcpp::Rcout << "Building interpolant for fine point " << row << std::endl;
-
       NumericT row_sum = 0;
       NumericT row_coarse_sum = 0;
       NumericT diag = 0;
@@ -793,9 +988,85 @@ void amg_interpol_direct(compressed_matrix<NumericT> const & A,
       }
 
       NumericT temp_res = -row_sum/(row_coarse_sum*diag);
-      //Rcpp::Rcout << "row_sum: " << row_sum << ", row_coarse_sum: " << row_coarse_sum << ", diag: " << diag << std::endl;
 
-      if (std::fabs(temp_res) > 1e-2 * std::fabs(diag))
+      if (std::fabs(temp_res) > threshold * std::fabs(diag))
+      {
+        unsigned int num_nonzeros_in_row = 0;
+
+        unsigned int row_A_start = A_row_buffer[row];
+        unsigned int row_A_end   = A_row_buffer[row + 1];
+        unsigned int const *influence_iter = influences_id_ptr + influences_row_ptr[row];
+        unsigned int const *influence_end  = influences_id_ptr + influences_row_ptr[row + 1];
+
+        for (unsigned int index = row_A_start; index < row_A_end; ++index)
+        {
+          unsigned int col = A_col_buffer[index];
+          if (point_types_ptr[col] != viennacl::linalg::detail::amg::amg_level_context::POINT_TYPE_COARSE)
+            continue;
+          NumericT value = A_elements[index];
+
+          // Advance to correct influence metric:
+          while (influence_iter != influence_end && *influence_iter < col)
+            ++influence_iter;
+
+          if (influence_iter != influence_end && *influence_iter == col && std::fabs(value) > 0)
+            ++num_nonzeros_in_row;
+        }
+
+        interpolation_helper[row] = temp_res;
+        P_nonzero_helper[row] = num_nonzeros_in_row; // as many nonzero entries as there are coarse points
+      }
+      else
+      {
+        interpolation_helper[row] = 0; // this ensures that the above if-statement is never valid if tested against zero
+        P_nonzero_helper[row] = 0;
+      }
+    }
+  }
+  P_nonzero_helper[A.size1()] = 0;
+
+  viennacl::vector_base<unsigned int> P_nonzero_helper_wrapper(P_nonzero_helper, viennacl::MAIN_MEMORY, A.size1() + 1);
+  viennacl::linalg::exclusive_scan(P_nonzero_helper_wrapper);
+
+
+  //
+  // Step 2: Populate P
+  //
+  P = compressed_matrix<NumericT>(A.size1(), amg_context.num_coarse_, P_nonzero_helper[A.size1()], viennacl::traits::context(A));
+
+  NumericT     * P_elements   = viennacl::linalg::host_based::detail::extract_raw_pointer<NumericT>(P.handle());
+  unsigned int * P_row_buffer = viennacl::linalg::host_based::detail::extract_raw_pointer<unsigned int>(P.handle1());
+  unsigned int * P_col_buffer = viennacl::linalg::host_based::detail::extract_raw_pointer<unsigned int>(P.handle2());
+
+
+  // Iterate over all points to build the interpolation matrix row-by-row
+  // Interpolation for coarse points is immediate using '1'.
+  // Interpolation for fine points is set up via corresponding row weights (cf. Yang paper, p. 14)
+#ifdef VIENNACL_WITH_OPENMP
+  #pragma omp parallel for
+#endif
+  for (long row2=0; row2<static_cast<long>(A.size1()); ++row2)
+  {
+    unsigned int row = static_cast<unsigned int>(row2);
+
+    unsigned int row_start = P_nonzero_helper[row];
+    P_row_buffer[row] = row_start;
+
+    if (point_types_ptr[row] == viennacl::linalg::detail::amg::amg_level_context::POINT_TYPE_COARSE)
+    {
+      P_col_buffer[row_start] = coarse_id_ptr[row];
+      P_elements[row_start]   = NumericT(1);
+    }
+    else if (point_types_ptr[row] == viennacl::linalg::detail::amg::amg_level_context::POINT_TYPE_FINE)
+    {
+      NumericT temp_res = interpolation_helper[row];
+
+      unsigned int row_A_start = A_row_buffer[row];
+      unsigned int row_A_end   = A_row_buffer[row + 1];
+      unsigned int const *influence_iter = influences_id_ptr + influences_row_ptr[row];
+      unsigned int const *influence_end  = influences_id_ptr + influences_row_ptr[row + 1];
+
+      if (std::fabs(temp_res) > 0)
       {
         // Iterate over all strongly influencing points to build the interpolant
         influence_iter = influences_id_ptr + influences_row_ptr[row];
@@ -810,10 +1081,11 @@ void amg_interpol_direct(compressed_matrix<NumericT> const & A,
           while (influence_iter != influence_end && *influence_iter < col)
             ++influence_iter;
 
-          if (influence_iter != influence_end && *influence_iter == col)
+          if (influence_iter != influence_end && *influence_iter == col && std::fabs(value) > 0)
           {
-            //Rcpp::Rcout << " Setting entry "  << temp_res * value << " at " << coarse_id_ptr[col] << " for point " << col << std::endl;
-            P_setup_row[coarse_id_ptr[col]] = temp_res * value;
+            P_col_buffer[row_start] = coarse_id_ptr[col];
+            P_elements[row_start]   = temp_res * value;
+            ++row_start;
           }
         }
       }
@@ -824,13 +1096,146 @@ void amg_interpol_direct(compressed_matrix<NumericT> const & A,
     else
       throw std::runtime_error("Logic error in direct interpolation: Point is neither coarse-point nor fine-point!");
   }
+  P_row_buffer[A.size1()] = P_nonzero_helper[A.size1()];
+
+  P.generate_row_block_information();
+
+  free(P_nonzero_helper);
+  free(interpolation_helper);
+}
+
+
+/** @brief Extended+i interpolation. Multi-threaded! (VIENNACL_AMG_INTERPOL_EXTENDED_I)
+ *
+ * See "Distance-Two Interpolation for Parallel Algebraic Multigrid" by H. De Sterck et al. (2007) for description
+ *
+ * @param A            Operator matrix
+ * @param P            Prolongation matrix
+ * @param amg_context  AMG hierarchy datastructures
+ * @param tag          AMG preconditioner tag
+*/
+template<typename NumericT>
+void amg_interpol_extended_i_mis1(compressed_matrix<NumericT> const & A,
+                                compressed_matrix<NumericT> & P,
+                                viennacl::linalg::detail::amg::amg_level_context & amg_context,
+                                viennacl::linalg::amg_tag & tag)
+{
+  (void)tag;
+
+  NumericT     const * A_elements   = viennacl::linalg::host_based::detail::extract_raw_pointer<NumericT>(A.handle());
+  unsigned int const * A_row_buffer = viennacl::linalg::host_based::detail::extract_raw_pointer<unsigned int>(A.handle1());
+  unsigned int const * A_col_buffer = viennacl::linalg::host_based::detail::extract_raw_pointer<unsigned int>(A.handle2());
+
+  unsigned int *point_types_ptr       = viennacl::linalg::host_based::detail::extract_raw_pointer<unsigned int>(amg_context.point_types_.handle());
+  unsigned int *influences_row_ptr    = viennacl::linalg::host_based::detail::extract_raw_pointer<unsigned int>(amg_context.influence_jumper_.handle());
+  unsigned int *influences_id_ptr     = viennacl::linalg::host_based::detail::extract_raw_pointer<unsigned int>(amg_context.influence_ids_.handle());
+  unsigned int *coarse_id_ptr         = viennacl::linalg::host_based::detail::extract_raw_pointer<unsigned int>(amg_context.coarse_id_.handle());
+
+  P.resize(A.size1(), amg_context.num_coarse_, false);
+
+  std::vector<std::map<unsigned int, NumericT> > P_setup(A.size1());
+
+  // Iterate over all points to build the interpolation matrix row-by-row
+  // Interpolation for coarse points is immediate using '1'.
+  // Interpolation for fine points is set up via (4.10) and (4.11) in "Distance-Two Interpolation for Parallel Algebraic Multigrid"
+#ifdef VIENNACL_WITH_OPENMP
+  #pragma omp parallel for
+#endif
+  for (long row2=0; row2<static_cast<long>(A.size1()); ++row2)
+  {
+    unsigned int row = static_cast<unsigned int>(row2);
+    std::map<unsigned int, NumericT> & P_setup_row = P_setup[row]; // w_ij (first assembled as w_ij * \tilde{a}_ii)
+
+    if (point_types_ptr[row] == viennacl::linalg::detail::amg::amg_level_context::POINT_TYPE_COARSE)
+    {
+      P_setup_row[coarse_id_ptr[row]] = NumericT(1);
+    }
+    else if (point_types_ptr[row] == viennacl::linalg::detail::amg::amg_level_context::POINT_TYPE_FINE)
+    {
+      // TODO: Add consideration of strong connections (currently all connections are considered strong)
+      //unsigned int const *influence_iter = influences_id_ptr + influences_row_ptr[row];
+      //unsigned int const *influence_end  = influences_id_ptr + influences_row_ptr[row + 1];
+
+      NumericT atilde_ii = 0;
+
+      unsigned int row_A_start = A_row_buffer[row];
+      unsigned int row_A_end   = A_row_buffer[row + 1];
+      for (unsigned int index = row_A_start; index < row_A_end; ++index)
+      {
+        unsigned int col = A_col_buffer[index];
+
+        if (col == row)
+        {
+          atilde_ii += A_elements[index];  // a_ii contribution to \tilde{a}_ii
+        }
+        else if (point_types_ptr[col] == viennacl::linalg::detail::amg::amg_level_context::POINT_TYPE_COARSE)
+        {
+          P_setup_row[coarse_id_ptr[col]] += A_elements[index]; // a_ij
+
+          // contributions from fine nodes influencing this coarse node follow below
+        }
+        else if (point_types_ptr[col] == viennacl::linalg::detail::amg::amg_level_context::POINT_TYPE_FINE) // reach out for coarse nodes at distance 2
+        {
+          NumericT aik = A_elements[index];
+          NumericT akl = 0;
+
+          unsigned int distance_2_start = A_row_buffer[col];
+          unsigned int distance_2_end   = A_row_buffer[col + 1];
+
+          // determine akl expression in denominator first:
+          for (unsigned int index_2 = distance_2_start; index_2 < distance_2_end; ++index_2)
+          {
+            unsigned int col_2 = A_col_buffer[index_2];
+
+            if (col_2 == row || point_types_ptr[col_2] == viennacl::linalg::detail::amg::amg_level_context::POINT_TYPE_COARSE)
+              akl += A_elements[index_2];
+          }
+
+          // compute numerator and denominator of w_ij
+          for (unsigned int index_2 = distance_2_start; index_2 < distance_2_end; ++index_2)
+          {
+            unsigned int col_2 = A_col_buffer[index_2];
+            NumericT akj = 0;
+            NumericT aki = 0;
+
+            if (point_types_ptr[col_2] == viennacl::linalg::detail::amg::amg_level_context::POINT_TYPE_COARSE)
+              akj  = A_elements[index_2];
+
+            if (col_2 == row)
+              aki  = A_elements[index_2];
+
+            if (std::fabs(akj) > 0)
+              P_setup_row[coarse_id_ptr[col_2]] += aik * akj / akl;
+            if (std::fabs(aki) > 0)
+              atilde_ii += aik * aki / akl;
+          }
+        }
+      }
+
+      if (atilde_ii <= 0 && atilde_ii >= 0)
+        throw std::runtime_error("amg_interpol_extended_i_mis1(): Denominator zero, interpolation invalid!");
+
+      // now divide the entries in P_setup_row by \tilde{a}_ii to form w_ij:
+      // Note: A minus sign is missing in Eq. 4.10 in "Distance-Two Interpolation for Parallel Algebraic Multigrid".
+      // Compare with Eq. 4.6, a minus sign appears there
+      for (typename std::map<unsigned int, NumericT>::iterator it  = P_setup_row.begin();
+                                                               it != P_setup_row.end();
+                                                             ++it)
+        it->second /= NumericT(-1.0) * atilde_ii;
+    }
+    else
+      throw std::runtime_error("Logic error in direct interpolation: Point is neither coarse-point nor fine-point!");
+  }
 
   // TODO: P_setup can be avoided without sacrificing parallelism.
   viennacl::tools::sparse_matrix_adapter<NumericT> P_adapter(P_setup, P.size1(), P.size2());
   viennacl::copy(P_adapter, P);
+
 }
 
 
+
+
 /** @brief AG (aggregation based) interpolation. Multi-Threaded! (VIENNACL_INTERPOL_AG)
  *
  * @param A            Operator matrix
@@ -984,41 +1389,87 @@ void amg_transpose(compressed_matrix<NumericT> const & A,
   unsigned int * B_row_buffer = viennacl::linalg::host_based::detail::extract_raw_pointer<unsigned int>(B.handle1());
   unsigned int * B_col_buffer = viennacl::linalg::host_based::detail::extract_raw_pointer<unsigned int>(B.handle2());
 
-  // prepare uninitialized B_row_buffer:
-  for (std::size_t i = 0; i < B.size1(); ++i)
-    B_row_buffer[i] = 0;
+#ifdef VIENNACL_WITH_OPENMP
+  int threads = 8;
+#else
+  int threads = 1;
+#endif
+  std::size_t scratchpad_size = threads * (B.size1()+1); //column-oriented matrix with 'threads' columns
+  unsigned int * scratchpad = (unsigned int *)malloc(sizeof(unsigned int) * scratchpad_size);
+
+  // prepare uninitialized scratchpad:
+#ifdef VIENNACL_WITH_OPENMP
+  #pragma omp parallel for
+#endif
+  for (std::size_t i = 0; i < scratchpad_size; ++i)
+    scratchpad[i] = 0;
 
   //
   // Stage 1: Compute pattern for B
   //
+#ifdef VIENNACL_WITH_OPENMP
+  #pragma omp parallel for num_threads(threads)
+#endif
   for (std::size_t row = 0; row < A.size1(); ++row)
   {
+    unsigned int thread_id = 0;
+#ifdef VIENNACL_WITH_OPENMP
+    thread_id = omp_get_thread_num();
+#endif
+
     unsigned int row_start = A_row_buffer[row];
     unsigned int row_stop  = A_row_buffer[row+1];
 
     for (unsigned int nnz_index = row_start; nnz_index < row_stop; ++nnz_index)
-      B_row_buffer[A_col_buffer[nnz_index]] += 1;
+      scratchpad[thread_id * (B.size1()+1) + A_col_buffer[nnz_index]] += 1;
   }
 
-  // Bring row-start array in place using inclusive-scan:
-  unsigned int offset = B_row_buffer[0];
-  B_row_buffer[0] = 0;
-  for (std::size_t row = 1; row < B.size1(); ++row)
+#ifdef VIENNACL_WITH_OPENMP
+  #pragma omp parallel for
+#endif
+  for (std::size_t row = 0; row < B.size1(); ++row)
   {
-    unsigned int tmp = B_row_buffer[row];
-    B_row_buffer[row] = offset;
-    offset += tmp;
+    unsigned int offset = scratchpad[row];
+    for (std::size_t i = 1; i<threads; ++i)
+    {
+      unsigned int tmp = scratchpad[i*(B.size1()+1) + row];
+      scratchpad[i*(B.size1()+1) + row] = offset;
+      offset += tmp;
+    }
+    scratchpad[row] = offset;
   }
-  B_row_buffer[B.size1()] = offset;
 
   //
-  // Stage 2: Fill with data
+  // Stage 2: Bring row-start array in place using exclusive-scan:
   //
+  viennacl::vector_base<unsigned int> helper_vec(scratchpad, viennacl::MAIN_MEMORY, B.size1()+1);
+  viennacl::linalg::host_based::exclusive_scan(helper_vec, helper_vec);
 
-  std::vector<unsigned int> B_row_offsets(B.size1()); //number of elements already written per row
+  // propagate offsets and copy CSR datastructure over to B:
+#ifdef VIENNACL_WITH_OPENMP
+  #pragma omp parallel for
+#endif
+  for (std::size_t row = 0; row < B.size1(); ++row)
+  {
+    unsigned int row_offset = scratchpad[row];
+    B_row_buffer[row] = row_offset;
+    for (std::size_t i = 1; i<threads; ++i)
+      scratchpad[i*(B.size1()+1) + row] += row_offset;
+  }
+  B_row_buffer[B.size1()] = scratchpad[B.size1()];
 
+  //
+  // Stage 3: Fill with data
+  //
+#ifdef VIENNACL_WITH_OPENMP
+  #pragma omp parallel for num_threads(threads)
+#endif
   for (std::size_t row = 0; row < A.size1(); ++row)
   {
+    unsigned int thread_id = 0;
+#ifdef VIENNACL_WITH_OPENMP
+    thread_id = omp_get_thread_num();
+#endif
     //Rcpp::Rcout << "Row " << row << ": ";
     unsigned int row_start = A_row_buffer[row];
     unsigned int row_stop  = A_row_buffer[row+1];
@@ -1026,16 +1477,18 @@ void amg_transpose(compressed_matrix<NumericT> const & A,
     for (unsigned int nnz_index = row_start; nnz_index < row_stop; ++nnz_index)
     {
       unsigned int col_in_A = A_col_buffer[nnz_index];
-      unsigned int B_nnz_index = B_row_buffer[col_in_A] + B_row_offsets[col_in_A];
+      unsigned int array_index = thread_id * (B.size1()+1) + col_in_A;
+      unsigned int B_nnz_index = scratchpad[array_index];
+      scratchpad[array_index] += 1;
       B_col_buffer[B_nnz_index] = static_cast<unsigned int>(row);
       B_elements[B_nnz_index] = A_elements[nnz_index];
-      ++B_row_offsets[col_in_A];
-      //B_temp.at(A_col_buffer[nnz_index])[row] = A_elements[nnz_index];
     }
   }
 
   // Step 3: Make datastructure consistent (row blocks!)
   B.generate_row_block_information();
+
+  free(scratchpad);
 }
 
 /** Assign sparse matrix A to dense matrix B */
