@@ -115,9 +115,9 @@ public:
     //Rcpp::Rcout << "Current device id in context: " << current_device_id_ << std::endl;
     return devices_[current_device_id_];
   }
-  
+
   /** @brief Returns the current device index */
-  int current_device_id() const
+  vcl_size_t current_device_id() const
   {
     return current_device_id_;
   }
@@ -205,7 +205,7 @@ public:
 #if defined(VIENNACL_DEBUG_ALL) || defined(VIENNACL_DEBUG_CONTEXT)
     Rcpp::Rcout << "ViennaCL: Creating memory of size " << size << " for context " << h_ << " (unsafe, returning cl_mem directly)" << std::endl;
 #endif
-    if (ptr)
+    if (ptr && !(flags & CL_MEM_USE_HOST_PTR))
       flags |= CL_MEM_COPY_HOST_PTR;
     cl_int err;
     cl_mem mem = clCreateBuffer(h_.get(), flags, size, ptr, &err);
